from __future__ import print_function

import sys
import threading
try:
    import queue
except ImportError:
    import Queue as queue

from . import logs
from . import utils
from . import sockio
from . import introspect

OUTPUT_MODES = ('local', 'remote', 'mirror')

log = logs.get(__name__)

def serve(address, locs=None):
    Service(locs).serve(address)

def connect(address, locs=None):
    Service(locs).connect(address)

class Service(object):
    def __init__(self, locs=None, filename=None):
        self._locals = {}
        self._is_evaluating = False

        self._stop = threading.Event()
        self._events = queue.Queue()
        self.code_queue = queue.Queue()

        self._code = introspect.Code(self._locals, filename)
        self._ctx = Context(self, locs or {})

        self.reset()

    ## execution ##

    def connect(self, address):
        def _connect(svc, address):
            try:
                svc.handle(sockio.connect(address))
            except sockio.error as e:
                log.error('connection failed: %s', e)

        utils.start_thread(_connect, self, address)

        self.run()

    def serve(self, address):
        sockio.start_server(address, self.handle)
        self.run()

    def run(self):
        q = self.code_queue

        while True:
            data = q.get()
            self.evaluate(**data)

    ## commands ##

    def evaluate(self, source, notify=True):
        self._is_evaluating = True
        try:
            self._code.evaluate(source)
        finally:
            self._is_evaluating = False
            if notify:
                self.add_event('done')

    def interrupt(self):
        self._code.interrupt()

    def complete(self, prefix):
        matches = self._code.complete(prefix)
        self.add_event('completion', matches=matches)

    def events(self):
        stop = self._stop
        q = self._events

        self.add_event('start', version=sys.version)

        while not stop.is_set():
            try:
                yield q.get(timeout=1)
            except queue.Empty:
                yield

    ## handlers ##

    def handle(self, sock):
        self.add_event('start', version=sys.version)

        t_evt = utils.start_thread(self.handle_events, sock)
        t_cmd = utils.start_thread(self.handle_commands, sock)

        t_evt.join()
        t_cmd.join()

    def handle_events(self, sock):
        stop = self._stop
        events = self._events

        try:
            while not stop.is_set():
                try:
                    event = events.get(timeout=1)
                except queue.Empty:
                    sock.sendmsg(None)
                else:
                    sock.sendmsg(event)
        except sockio.error as e:
            log.warning('handle_events error: %s', repr(e))

    def handle_commands(self, sock):
        stop = self._stop

        try:
            while not stop.is_set():
                msg = sock.recvmsg()

                cmd = msg['cmd']
                data = msg.get('data')
                log.debug('cmd: %s%s', cmd, ': ' + repr(data) if data else '')

                if cmd == 'evaluate':
                    if self._is_evaluating:
                        self.recv_input(data + '\n')
                    else:
                        self.code_queue.put(data)
                elif cmd == 'interrupt':
                    self.interrupt()
                elif cmd == 'complete':
                    self.complete(data)
                else:
                    log.error('unknown command: %s', cmd)

        except sockio.ReceiveInterrupted as e:
            log.warning('handle_commands error: %s', repr(e))

    ## utils ##

    def add_event(self, name, **data):
        if name == 'output':
            log.debug('out: %r', data['text'][:100])
        else:
            log.debug('evt: %s%s', name, (data or '') and ': ' + repr(data))

        event = {'evt': name, 'data': data}
        self._events.put(event)

    def recv_input(self, text):
        sys.stdin.write(text)

    def reset(self):
        self._code.reset()
        self._code.locals.update(self._ctx.env)
        self._result_count = 0

    def stop(self):
        self._stop.set()

class Context(object):
    """Exposes a limited API to the control terminal."""
    def __init__(self, service, loc):
        self.env = {
            '__name__': '__remote__',
            '__context__': self,
            }
        self.env.update(loc)

        self._service = service

        self._output_mode = None
        self.output_mode = 'remote'

    @property
    def output_mode(self):
        """Determines where interpreter output is displayed.

        There are 3 valid modes:

            local  - output is only displayed locally
            remote - output is only displayed remotely
            mirror - output is displayed both locally and remotely
        """
        return self._output_mode

    @output_mode.setter
    def output_mode(self, mode):
        svc = self._service

        if mode and mode == self._output_mode:
            return
        elif mode == 'local':
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__
        elif mode == 'remote':
            sys.stdout = introspect.OutputIO(svc)
            sys.stderr = introspect.OutputIO(svc)
        elif mode == 'mirror':
            sys.stdout = introspect.OutputIO(svc, sys.__stdout__)
            sys.stderr = introspect.OutputIO(svc, sys.__stderr__)
        else:
            err = 'output_mode must be one of: {}'
            raise ValueError(err.format(', '.join(OUTPUT_MODES)))
        self._output_mode = mode

    def reset(self):
        """Clears and resets the locals environment."""
        self._service.reset()

    def __str__(self):
        return "{}(output_mode='{}', env={})".format(
            self.__class__.__name__, self.output_mode, self.env,
            )



from qtpy import QtCore, QtWidgets

class InterpreterWidget(QtWidgets.QWidget):
    interpreter_changed = QtCore.Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setup()

    def setup(self):
        self.layout = QtWidgets.QFormLayout()
        self.setLayout(self.layout)

        self.python_exec_edit = QtWidgets.QLineEdit()
        self.layout.addRow('Python executable', self.python_exec_edit)

        self.setup_actions()

    def setup_actions(self):
        self.python_exec_edit.textChanged.connect(self.interpreter_changed)

    def set_python_exec(self, path):
        self.python_exec_edit.setText(path)



import io
import json
import errno
import socket
import struct
import threading

from . import logs
from .utils import start_thread

BACKLOG = socket.SOMAXCONN
CHUNK_SIZE = io.DEFAULT_BUFFER_SIZE

error = socket.error

log = logs.get(__name__)

def connect(address, timeout=None):
    log.debug('connecting: %s:%s', *address)
    sock = socket.create_connection(address, timeout)
    log.info('connected: %s:%s', *address)
    return SockIO(sock)

def start_server(address, handler, timeout=None, accept_timeout=1, backlog=None):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(accept_timeout)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(address)
    sock.listen(backlog or BACKLOG)

    host, port = sock.getsockname()
    log.info('listening: %s:%s', host, port)

    stop = threading.Event()
    t = start_thread(serve, sock, handler, stop, timeout)
    return (ServerThread(t, stop), (host, port))

def serve(sock, handler, stop, timeout=None):
    while not stop.is_set():
        try:
            s, addr = sock.accept()
        except socket.timeout:
            continue
        s.settimeout(timeout)

        log.info('connected: %s:%s', *addr)
        start_thread(handler, SockIO(s))

class ServerThread(object):
    def __init__(self, thread, stop):
        self._thread = thread
        self._stop = stop

    def stop(self):
        self._stop.set()

    def join(self):
        self._thread.join()

class SockIO(object):
    def __init__(self, sock, chunk_size=None):
        self._sock = sock
        self._chunk_size = chunk_size or CHUNK_SIZE

    def sendmsg(self, msg):
        data = json.dumps(msg).encode('utf8')
        self.send(data)

    def recvmsg(self):
        data = self.recv()
        return json.loads(data.decode('utf8'))

    def send(self, data):
        data_len = len(data)
        size = struct.pack('>I', data_len)
        self._sock.sendall(size)
        self._sock.sendall(data)

    def recv(self):
        return b''.join(self.recviter())

    def recviter(self):
        buf = b''.join(self.recvsize(4))
        data_len = struct.unpack('>I', buf)[0]
        for chunk in self.recvsize(data_len):
            yield chunk

    def recvsize(self, size):
        sock = self._sock

        pos = 0
        chunk_size = min(size, self._chunk_size)
        while pos < size:
            chunk = sock.recv(min(size-pos, chunk_size))
            if not chunk:
                raise ReceiveInterrupted()
            pos += len(chunk)
            yield chunk

    def settimeout(self, t):
        self._sock.settimeout(t)

    def close(self):
        try:
            self._sock.shutdown(socket.SHUT_RDWR)
        except (OSError, socket.error) as e:
            # ignore if not connected
            if e.errno not in (errno.ENOTCONN,):
                raise
        self._sock.close()

class SockIOError(Exception):
    pass

class ReceiveInterrupted(SockIOError):
    pass



import time
import shlex
import threading
import subprocess
import collections

from . import logs
from . import utils
from . import sockio

MODULE = 'telepythy'
TIMEOUT = 5

log = logs.get(__name__)

class Manager:
    def __init__(self, config, verbose=0, quiet=False):
        self._config = config
        self._verbose = verbose
        self._quiet = quiet

    def get_control(self, profile=None, connect=None, serve=None):
        command = None

        if profile is not None or (connect is None and serve is None):
            profile = profile or 'default'

            try:
                sec = self._config.profile[profile]
            except KeyError:
                # must be a command
                command = profile
            else:
                command = sec.get('command')
                connect = sec.get('connect')
                serve = sec.get('serve')

        if command is not None:
            return ProcessControl(('localhost', 0), command,
                self._verbose, self._quiet)

        elif connect is not None:
            addr = utils.parse_address(connect or utils.DEFAULT_ADDR)
            return ClientControl(addr)

        elif serve is not None:
            addr = utils.parse_address(serve or utils.DEFAULT_ADDR)
            return ServerControl(addr)

        assert False, 'invalid control init'

class Control(object):
    def __init__(self, address):
        self._address = address

        self._handlers = collections.defaultdict(set)

        self._work_socket = None
        self._work_event = threading.Event()
        self._events_thread = None
        self._events_stop = threading.Event()

        self._restart_lock = threading.Lock()

    def init(self):
        raise NotImplementedError('abstract')

    ## commands ##

    def evaluate(self, source, notify=True):
        self.get_work_proxy().evaluate(source, notify)

    def interrupt(self):
        self.get_work_proxy().interrupt()

    def complete(self, prefix):
        self.get_work_proxy().complete(prefix)

    ## events ##

    def register(self, event, handler):
        self._handlers[event].add(handler)

    def _events(self, sock):
        stop = self._events_stop
        stop.clear()

        handle = self._handle_event
        address = self._address

        proxy = ServiceProxy(sock)

        try:
            for event in proxy.events(stop):
                if stop.is_set():
                    break
                if event is None:
                    handle(None, address)
                    continue
                name = event['evt']
                handle(name, event)
        except Exception as e:
            if stop.is_set():
                return
            handle('error', repr(e))
            self.restart()

    def _handle_event(self, name, event=None):
        for handler in self._handlers.get(name, []):
            handler(event)

    ## proxies ##

    def get_work_proxy(self):
        if not self._work_socket:
            raise Exception('no connection')
        return ServiceProxy(self._work_socket)

    ## sockets ##

    def _set_sockets(self, sock):
        self._work_socket = sock
        self._work_event.set()
        self._events_thread = utils.start_thread(self._events, sock)

    def restart(self):
        with self._restart_lock:
            log.debug('restarting')
            self._restart()

    def _restart(self):
        self._work_event.clear()
        self._work_socket = None
        self._events_stop.set()
        self._events_thread = None

    def shutdown(self):
        self._events_stop.set()

class ClientControl(Control):
    def __init__(self, address):
        super().__init__(address)
        self._connect_stop = threading.Event()

    def init(self):
        self.connect()

    def connect(self):
        stop = self._connect_stop

        def connect(address):
            while not stop.is_set():
                try:
                    sock = sockio.connect(address)
                except sockio.error as e:
                    log.debug('connection failed: %s', e)
                    time.sleep(1)
                    continue
                else:
                    self._set_sockets(sock)
                    break

        utils.start_thread(connect, self._address)

    def _restart(self):
        super()._restart()
        self.connect()

    def shutdown(self):
        super().shutdown()
        self._connect_stop.set()

class ServerControl(Control):
    def __init__(self, address):
        super().__init__(address)
        self._server_thread = None

    def init(self):
        self.serve()

    def serve(self):
        # replace address in case a port was generated (port=0)
        self._server_thread, self._address = sockio.start_server(
            self._address, self._set_sockets)

    def shutdown(self):
        if self._server_thread:
            self._server_thread.stop()
        super().shutdown()

class ProcessControl(ServerControl):
    def __init__(self, address, command=None, verbose=0, quiet=False, kill_timeout=None):
        super().__init__(address)

        self._proc = None

        self._command = command
        self._verbose = verbose
        self._quiet = quiet
        self._timeout = TIMEOUT if kill_timeout is None else kill_timeout

    def init(self):
        self.serve()
        self.start()

    def start(self):
        if self._proc:
            return

        cmd = shlex.split(self._command) + ['-m', MODULE]
        if not self._quiet:
            cmd.extend(['-v'] * self._verbose)
        cmd.extend(['-c', '{}:{}'.format(*self._address)])

        log.debug('starting process: %s', cmd)
        self._proc = subprocess.Popen(cmd)

    def stop(self):
        proc = self._proc
        if not proc:
            return
        log.debug('stopping process')

        proc.terminate()
        try:
            proc.wait(self._timeout)
        except subprocess.TimeoutExpired:
            log.warning('terminate timed out. killing ...')
            proc.kill()
        proc.wait()

        self._proc = None

    def _restart(self):
        self.stop()
        super()._restart()
        self.start()

    def shutdown(self):
        super().shutdown()
        self.stop()

class ServiceProxy(object):
    def __init__(self, sock):
        self._sock = sock

    def evaluate(self, source, notify=True):
        self._sendcmd('evaluate', {'source': source, 'notify': notify})

    def interrupt(self):
        self._sendcmd('interrupt')

    def complete(self, prefix):
        self._sendcmd('complete', prefix)

    def events(self, stop):
        sock = self._sock

        while not stop.is_set():
            event = sock.recvmsg()

            if event:
                name = event['evt']
                if name == 'done':
                    log.debug('evt: done')
                elif name == 'output':
                    log.debug('out: %r', event['data']['text'][:100])
                else:
                    data = event.get('data', '')
                    data = data and ': ' + repr(data)[:100]
                    log.debug('evt: %s%s', event['evt'], data)

            yield event

    def _sendcmd(self, cmd, data=None):
        msg = {'cmd': cmd}
        if data is not None:
            msg['data'] = data
        data = (data or '') and ': ' + repr(data)
        log.debug('cmd: %s%s', cmd, data)
        self._sock.sendmsg(msg)



import re
from qtpy import QtGui, QtWidgets

rx_indent = re.compile('^[ ]*')

class BlockIterator:
    def __init__(self, document, first=None, last=None):
        self.document = document
        self.first = first
        self.last = last

    def __iter__(self):
        last = self.last
        block = self.first or self.document.firstBlock()
        while block.isValid():
            yield block
            if block == last:
                return
            block = block.next()

    def __reversed__(self):
        last = self.last
        block = self.first or self.document.lastBlock()
        while block.isValid():
            yield block
            if block == last:
                return
            block = block.previous()

class TextDocument(QtGui.QTextDocument):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setDocumentLayout(QtWidgets.QPlainTextDocumentLayout(self))

    def blocks(self, first=None, last=None):
        return BlockIterator(self, first, last)

    def blockIndentation(self, block):
        match = rx_indent.match(block.text())
        return match.end(0) - match.start(0) if match else 0



from qtpy import QtWidgets

PS1 = '>>> '
PS2 = '... '

class OutputEdit(QtWidgets.QPlainTextEdit):
    def append(self, text):
        cur = self.textCursor()
        cur.movePosition(cur.End)
        cur.insertText(text)
        self.scroll_to_bottom()

    def append_session(self, version):
        cur = self.textCursor()
        cur.movePosition(cur.End)
        if self.blockCount() > 1:
            cur.insertText('\n')
        tpl = '<div style="background: darkred;">{}</p><p></p>'
        cur.insertHtml(tpl.format(version))
        self.append_prompt()

    def append_prompt(self, prompt=PS1):
        self.append(prompt)

    def append_source(self, source):
        if not source:
            return

        cur = self.textCursor()
        cur.movePosition(cur.End)
        insert = cur.insertText
        lines = source.splitlines()
        insert(lines[0] + '\n')

        # TODO: remove trailing empty lines

        for line in lines[1:]:
            insert(PS2)
            insert(line + '\n')

        self.scroll_to_bottom()

    def extract_source(self):
        def collect():
            text = self.toPlainText()
            for line in text.splitlines():
                if line.startswith(PS1):
                    yield line[len(PS1):]
                elif line.startswith(PS2):
                    yield line[len(PS2):]
        return '\n'.join(collect())

    def scroll_to_bottom(self):
        scroll = self.verticalScrollBar()
        scroll.setValue(scroll.maximum())



import collections

from qtpy.QtCore import Qt
from qtpy import QtCore, QtGui, QtWidgets

from pygments.lexers import PythonConsoleLexer

from .. import logs
from .. import utils

from .source_edit import SourceEdit
from .output_edit import OutputEdit
from .style_widget import StyleWidget
from .highlighter import PygmentsHighlighter

log = logs.get(__name__)

class Window(QtWidgets.QMainWindow):
    output_started = QtCore.Signal(str)
    output_stopped = QtCore.Signal()
    output_received = QtCore.Signal(str)
    completion_received = QtCore.Signal(list)
    status_connected = QtCore.Signal(tuple)
    status_disconnected = QtCore.Signal(str)

    def __init__(self, config, manager, profile):
        super().__init__()

        self._manager = manager
        self._control = None
        self._profiles = {}

        self.setup()
        self.config(config, profile)
        self.set_profile(profile)

        self._connected = False
        self._history_result = collections.OrderedDict()

        self._set_disconnected(force=True)

    ## config ##

    def config(self, config, profile):
        self._config = config

        # style
        sec = config.style
        self.style_chooser.set_app_style(sec.app)
        self.style_chooser.set_output_style(sec.output)
        self.style_chooser.set_source_style(sec.source)

        # menus
        view_menu = config.window.view.menu
        self.menuBar().setVisible(view_menu)
        self.action_toggle_menu.setChecked(view_menu)

        self.action_toggle_source_title.setChecked(False)

        # profiles
        menu = self.profile_menu
        group = QtWidgets.QActionGroup(menu)
        for name in config.profile:
            action = group.addAction(name)
            action.setCheckable(True)
            action.setChecked(name == profile)
            menu.addAction(action)
            self._profiles[name] = action

        self.resize(*config.window.size)

    ## setup ##

    def setup(self):
        self.setup_actions()
        self.setup_output_edit()
        self.setup_source_edit()
        self.setup_style_widget()
        self.setup_menus()
        self.setup_menubar()
        self.setup_statusbar()
        self.setup_signals()

        self.source_edit.setFocus()

    def setup_actions(self):
        self.action_quit = QtWidgets.QAction('Quit')
        self.action_quit.setShortcut('Ctrl+q')
        self.addAction(self.action_quit)

        self.action_interrupt = QtWidgets.QAction('Interrupt')
        self.action_interrupt.setShortcut('Ctrl+c')
        self.addAction(self.action_interrupt)

        self.action_restart = QtWidgets.QAction('Restart')
        self.action_restart.setShortcut('Ctrl+F6')
        self.addAction(self.action_restart)

        self.action_toggle_menu = QtWidgets.QAction('Menu')
        self.action_toggle_menu.setCheckable(True)
        self.action_toggle_menu.setChecked(True)

        self.action_toggle_source_title = QtWidgets.QAction('Source Titlebar')
        self.action_toggle_source_title.setCheckable(True)
        self.action_toggle_source_title.setChecked(True)

    def setup_output_edit(self):
        self.output_edit = OutputEdit()
        self.output_edit.setFont(QtGui.QFont('Fira Mono', 13))
        self.output_edit.setReadOnly(True)
        self.output_edit.setWordWrapMode(QtGui.QTextOption.WrapAnywhere)

        self.output_highlighter = PygmentsHighlighter(
            self.output_edit.document(), PythonConsoleLexer())

        palette = self.output_edit.palette()
        palette.setColor(QtGui.QPalette.Base, '#333')
        palette.setColor(QtGui.QPalette.Text, Qt.white)
        self.output_edit.setPalette(palette)

        self.setCentralWidget(self.output_edit)

    def setup_source_edit(self):
        self.source_edit = SourceEdit()
        self.source_edit.setFont(QtGui.QFont('Fira Mono', 13))

        self.source_highlighter = PygmentsHighlighter(
            self.source_edit.document())

        palette = self.source_edit.palette()
        palette.setColor(QtGui.QPalette.Base, '#333')
        palette.setColor(QtGui.QPalette.Text, Qt.white)
        self.source_edit.setPalette(palette)

        self.source_dock = QtWidgets.QDockWidget('source')
        self.source_dock.setWidget(self.source_edit)

        self.addDockWidget(Qt.BottomDockWidgetArea, self.source_dock)

    def setup_style_widget(self):
        self.style_chooser = StyleWidget()

        self.style_dock = QtWidgets.QDockWidget('Styles')
        self.style_dock.setWidget(self.style_chooser)
        self.style_dock.setVisible(False)

        self.addDockWidget(Qt.RightDockWidgetArea, self.style_dock)

    def setup_menus(self):
        self.main_menu = QtWidgets.QMenu('Main', self)
        self.main_menu.addAction(self.action_restart)
        self.main_menu.addAction(self.action_quit)

        self.view_menu = QtWidgets.QMenu('View', self)
        self.view_menu.addAction(self.action_toggle_menu)
        self.view_menu.addAction(self.style_dock.toggleViewAction())
        self.view_menu.addAction(self.action_toggle_source_title)

        self.profile_menu = QtWidgets.QMenu('Profile', self)

        self.status_menu = QtWidgets.QMenu()
        self.status_menu.addMenu(self.main_menu)
        self.status_menu.addMenu(self.view_menu)
        self.status_menu.addMenu(self.profile_menu)

    def setup_menubar(self):
        bar = self.menuBar()
        bar.addMenu(self.main_menu)
        bar.addMenu(self.view_menu)
        bar.addMenu(self.profile_menu)

    def setup_statusbar(self):
        bar = self.statusBar()

        icon = QtGui.QIcon(':connected')
        self._status_pixmap_connected = icon.pixmap(16)
        icon = QtGui.QIcon(':disconnected')
        self._status_pixmap_disconnected = icon.pixmap(16)

        self.menu_button = QtWidgets.QPushButton()
        self.menu_button.setIcon(QtGui.QIcon(':menu'))
        self.menu_button.setMenu(self.status_menu)
        self.menu_button.setStyleSheet('::menu-indicator{ image: none; }')
        self.menu_button.setFlat(True)
        bar.addWidget(self.menu_button)

        self.status_label = QtWidgets.QLabel()
        bar.addPermanentWidget(self.status_label)

        self.status_icon = QtWidgets.QLabel()
        bar.addPermanentWidget(self.status_icon)

        self.profile_button = QtWidgets.QPushButton('default')
        self.profile_button.setMenu(self.profile_menu)
        self.profile_button.setStyleSheet('::menu-indicator{ image: none; }')
        self.profile_button.setFlat(True)
        bar.addPermanentWidget(self.profile_button)

    def setup_signals(self):
        self.action_quit.triggered.connect(self.close)
        self.action_interrupt.triggered.connect(self.interrupt)
        self.action_restart.triggered.connect(self.restart)
        self.action_toggle_menu.toggled.connect(self.menuBar().setVisible)

        def source_toggle(checked):
            w = None if checked else QtWidgets.QWidget(self.source_dock)
            self.source_dock.setTitleBarWidget(w)
        self.action_toggle_source_title.toggled.connect(source_toggle)

        self.profile_menu.triggered.connect(
            lambda action: self.set_profile(action.text()))
        self.profile_menu.triggered.connect(
            lambda action: self.profile_button.setText(action.text()))

        self.source_edit.evaluation_requested.connect(self.evaluate)
        self.source_edit.completion_requested.connect(self.complete)

        self.output_started.connect(self.start_session)
        self.output_stopped.connect(self.output_edit.append_prompt)
        self.output_received.connect(self.output_edit.append)

        self.completion_received.connect(self.source_edit.show_completer)

        self.status_connected.connect(self._set_connected)
        self.status_disconnected.connect(self._set_disconnected)

        self.style_chooser.output_style_changed.connect(
            self.output_highlighter.set_style)
        self.style_chooser.source_style_changed.connect(
            self.source_highlighter.set_style)

    ## events ##

    def contextMenuEvent(self, event):
        self.view_menu.exec_(event.globalPos())

    def closeEvent(self, event):
        self._control.shutdown()

    ## actions ##

    def set_profile(self, name):
        action = self._profiles[name]

        if self._control:
            self._control.shutdown()

        self._control = ctl = self._manager.get_control(name)

        ctl.register(None, lambda address: self.status_connected.emit(address))
        def start(event):
            version = event['data']['version']
            self.output_started.emit(version)
        ctl.register('start', start)
        ctl.register('done', lambda _: self.output_stopped.emit())

        def output(event):
            text = event['data']['text']
            self.output_received.emit(text)
        ctl.register('output', output)

        def completion(event):
            matches = event['data']['matches']
            self.completion_received.emit(matches)
        ctl.register('completion', completion)

        def error(err):
            log.debug('totally normal events error: %s', err)
            self.status_disconnected.emit(err)
        ctl.register('error', error)

        ctl.init()

        self.profile_menu.setActiveAction(action)
        self.profile_button.setText(name)

    def stop(self):
        self._control.stop()

    def restart(self):
        self._control.restart()

    ## commands ##

    def start_session(self, version):
        self.output_edit.append_session(version)

        source = self._config.startup.source
        if source:
            self._control.evaluate(source, notify=False)

    def evaluate(self, source):
        try:
            self._control.evaluate(source)
        except Exception as e:
            log.debug('totally normal evaluate error: %s', e)
            self.status_disconnected.emit(str(e))
        else:
            self.output_edit.append_source(source)
            self.source_edit.next_cell()

    def interrupt(self):
        try:
            self._control.interrupt()
        except Exception as e:
            log.debug('totally normal interrupt error: %s', e)
            self.status_disconnected.emit(str(e))

    def complete(self, context):
        try:
            self._control.complete(context)
        except Exception as e:
            log.debug('totally normal complete error: %s', e)
            self.status_disconnected.emit(str(e))

    ## status ##

    def _set_connected(self, address):
        if self._connected:
            return
        self._connected = True

        msg = 'connected: {}:{}'.format(*address)
        self.status_label.setText(msg)
        self.status_icon.setPixmap(self._status_pixmap_connected)

    def _set_disconnected(self, error=None, force=False):
        if not self._connected and not force:
            return
        self._connected = False

        e = error and ': {}'.format(error) or ''
        msg = 'not connected{}'.format(e)
        self.status_label.setText(msg)
        self.status_icon.setPixmap(self._status_pixmap_disconnected)



import re
import contextlib

from qtpy.QtCore import Qt
from qtpy import QtCore, QtGui, QtWidgets

from . import document
from .history import History

COMPLETER_KEYS = frozenset([
    Qt.Key_Return, Qt.Key_Enter, Qt.Key_Escape, Qt.Key_Tab, Qt.Key_Backtab])

_rx_indent = re.compile(r'^(\s*)')

# match dotted identifiers
_rx_dotted_ident = re.compile(r'.*[^\w.](.*)')

def get_completion_context(line):
    # the '!' is a hack because I can't figure out how to match when there is
    # no non-word/dot char
    match = _rx_dotted_ident.match('!' + line)
    return match and match.group(1)

# required to let ctrl+c through (when no text is selected)
class EventFilter(QtCore.QObject):
    def eventFilter(self, obj, event):
        if isinstance(event, QtGui.QKeyEvent):
            key = event.key()
            mod = event.modifiers()
            ctrl = mod & Qt.ControlModifier

            if ctrl and key == Qt.Key_C:
                if not obj.textCursor().hasSelection():
                    return True

        return super().eventFilter(obj, event)

class SourceEdit(QtWidgets.QPlainTextEdit):
    evaluation_requested = QtCore.Signal(str)
    completion_requested = QtCore.Signal(str)

    def __init__(self):
        super().__init__()

        self._history = History()
        self._current = None

        self.installEventFilter(EventFilter(self))

        doc = document.TextDocument(self)
        self.setDocument(doc)

        self.completer_model = QtGui.QStandardItemModel()
        self.completer = QtWidgets.QCompleter(self.completer_model, self)
        self.completer.setWidget(self)
        self.completer.setCompletionMode(self.completer.PopupCompletion)
        self.completer.activated[str].connect(self.complete)

        self.textChanged.connect(self.refresh_completer)

    def show_completer(self, matches):
        if not matches:
            return

        completer = self.completer
        model = self.completer_model

        model.clear()
        for match in matches:
            item = QtGui.QStandardItem(match)
            model.appendRow(item)

        if completer.completionCount() == 1:
            self.complete(completer.currentCompletion())
            return

        # popup size
        popup = completer.popup()
        rect = self.cursorRect()
        scrollbar = popup.verticalScrollBar()
        rect.setWidth(popup.sizeHintForColumn(0) + scrollbar.sizeHint().width())

        completer.complete(rect)

    def refresh_completer(self):
        # if not self.completer.popup().isVisible():
        #     return
        ident = self.completion_context().split('.')[-1]
        self.completer.setCompletionPrefix(ident)

        # popup = self.completer.popup()
        # item = self.completer_model.item(0)
        # if item:
        #     popup.setCurrentIndex(item.index())

    def complete(self, match):
        print(match)
        ident = self.completion_context().split('.')[-1]
        text = match[len(ident):]

        self.insertPlainText(text)
        self.completer_model.clear()

    def completion_context(self):
        cur = self.textCursor()
        cur.movePosition(cur.StartOfLine, cur.KeepAnchor)
        sel = cur.selectedText()

        return get_completion_context(sel)

    def previous(self):
        if not self._history:
            return

        match = self.toPlainText()
        if self._current is None:
            self._current = match

        source = self._history.previous(match)
        if source:
            self.setPlainText(source)
            self.move_cursor_position(QtGui.QTextCursor.End)

    def next(self):
        if not self._history:
            return

        match = self.toPlainText()
        source = self._history.next(match)
        if source:
            self.setPlainText(source)
        else:
            self.setPlainText(self._current)
        self.move_cursor_position(QtGui.QTextCursor.End)

    def reset(self):
        # clear history browsing state
        self._history.reset()
        self._current = None

    def setPlainText(self, text):
        """Overridden to prevent clearing undo history."""
        self.clear()
        self.insertPlainText(text)

    def clear(self):
        cur = self.textCursor()
        cur.movePosition(cur.Start)
        cur.movePosition(cur.End, cur.KeepAnchor)
        cur.deleteChar()

    def next_cell(self):
        self.reset()

        source = self.toPlainText()
        self._history.append(source)

        self.clear()

    ## events ##

    def keyPressEvent(self, event):
        key = event.key()
        mod = event.modifiers()
        ctrl = mod & Qt.ControlModifier

        if self.completer.popup().isVisible():
            if key in COMPLETER_KEYS:
                event.ignore()
                return

        self.ensureCursorVisible()

        if key == Qt.Key_Return:
            # only evaluate when all of the following are true:
            # - there is only one line
            # - the text cursor is at the end of the text
            # - the last character is not a space
            # - the last non-space character is not a colon
            # evaluation can be forced by holding control
            cursor = self.textCursor()
            block = cursor.block()
            text = block.text()

            one_line = self.blockCount() == 1
            at_end = cursor.atEnd()
            no_space = text and text[-1] != ' '
            text_s = text.rstrip()
            new_scope = text_s and (text_s[-1] == ':')

            if ctrl or (one_line and at_end and no_space and not new_scope):
                source = self.toPlainText()
                if source:
                    self.evaluation_requested.emit(source)
                return
            else:
                spaces = self.block_indent(block)
                if new_scope:
                    spaces += 4
                self.insertPlainText('\n' + (' ' * spaces))
                return

        elif key == Qt.Key_Tab:
            cursor = self.textCursor()
            if not cursor.hasSelection():
                ctx = self.completion_context()
                if ctx:
                    self.completion_requested.emit(ctx)
                    return

            self.indent()
            return

        elif key == Qt.Key_Backtab:
            self.dedent()
            return

        elif ctrl and key == Qt.Key_Up:
            self.previous()
            return

        elif ctrl and key == Qt.Key_Down:
            self.next()
            return

        elif key == Qt.Key_Backspace:
            cursor = self.textCursor()
            text = cursor.block().text()[:cursor.positionInBlock()]
            if text.startswith(' ') and not text.strip(' '):
                cursor.movePosition(cursor.PreviousCharacter,
                    cursor.KeepAnchor, 4)
                cursor.deleteChar()
                return

        self.reset()
        super().keyPressEvent(event)

    ## utils ##

    def indent(self):
        doc = self.document()
        cursor = self.textCursor()
        block = cursor.block()
        indent = 4
        with cursor_edit(cursor):
            if cursor.hasSelection():
                start = cursor.selectionStart()
                end = cursor.selectionEnd()
                for block in reversed(doc.blocks(doc.findBlock(end))):
                    blockEnd = block.position() + block.length() - 1
                    if blockEnd < start:
                        break
                    offset = doc.blockIndentation(block) % indent
                    cursor.setPosition(block.position())
                    cursor.insertText(' ' * (indent - offset))
            else:
                offset = doc.blockIndentation(block) % indent
                cursor.setPosition(block.position())
                cursor.insertText(' ' * (indent - offset))

    def dedent(self):
        doc = self.document()
        indent = 4
        cursor = self.textCursor()
        with cursor_edit(cursor):
            if cursor.hasSelection():
                start = cursor.selectionStart()
                end = cursor.selectionEnd()
                blocks = reversed(doc.blocks(doc.findBlock(end)))
            else:
                start = 0
                blocks = [cursor.block()]

            for block in blocks:
                blockEnd = block.position() + block.length() - 1
                if blockEnd < start:
                    break

                blockIndent = doc.blockIndentation(block)
                offset = blockIndent % indent

                cursor.setPosition(block.position() + blockIndent)
                n = min(cursor.columnNumber(), indent - offset)
                cursor.movePosition(cursor.PreviousCharacter, cursor.KeepAnchor, n)
                cursor.removeSelectedText()

    def blocks(self, start=None):
        return BlockIterator(start or self.firstVisibleBlock())

    def block_indent(self, block):
        for prev in reversed(self.blocks(block)):
            spaces = _rx_indent.match(prev.text()).group(0)
            if spaces:
                return len(spaces)
        return 0

    def move_cursor_position(self, position):
        cursor = self.textCursor()
        cursor.movePosition(position)
        self.setTextCursor(cursor)

    def scroll_to_bottom(self):
        scroll = self.verticalScrollBar()
        scroll.setValue(scroll.maximum())

@contextlib.contextmanager
def cursor_edit(cursor):
    cursor.beginEditBlock()
    try:
        yield cursor
    finally:
        cursor.endEditBlock()

class BlockIterator:
    def __init__(self, start):
        self._start = start

    def __iter__(self):
        block = self._start
        while block.isValid():
            yield block
            block = block.next()

    def __reversed__(self):
        block = self._start
        while block.isValid():
            yield block
            block = block.previous()



class History(object):
    def __init__(self):
        self._history = []
        self._index = 0
        self._match = None

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, value):
        self._index = min(max(0, value), len(self._history))

    def append(self, value):
        hist = self._history
        value = value.strip()
        if value and value != (hist and hist[-1]):
            hist.append(value)
        self.reset()

    def first(self):
        self.index = 0
        return self._history[0]

    def last(self):
        self.index = len(self._history) - 1
        return self._history[-1]

    def previous(self, match=None):
        self.index -= 1
        i, value = self._search(match, range(self.index, -1, -1))
        self.index = i
        return value

    def next(self, match=None):
        self.index += 1
        i, value = self._search(match, range(self.index, len(self._history)))
        self.index = i
        return value

    def _search(self, match, it):
        if self._match is None:
            self._match = match
        match = self._match

        for i in it:
            value = self._history[i]
            if not match or value.startswith(match):
                return i, value

        return (self.index, None)

    def reset(self):
        self.index = len(self._history)
        self._match = None

    def __bool__(self):
        return bool(self._history)



import toml
import appdirs
from attrdict import AttrDict

DEFAULT_PATH = appdirs.user_config_dir('telepythy.cfg', False)
DEFAULT_INTERPRETER = 'python'

def init(path=None):
    path = path or DEFAULT_PATH

    defaults = AttrDict({
        'profile': {'default': {'command': DEFAULT_INTERPRETER}},
        'startup': {'source': ''},
        'style': {
            'app': 'qdarkstyle',
            'output': 'gruvbox-dark',
            'source': 'gruvbox-dark',
            },
        'window': {
            'size': [800, 800],
            'view': {
                'menu': True,
                },
            },
        })

    try:
        with open(path) as f:
            cfg = toml.load(f)
    except FileNotFoundError:
        cfg = {}

    cfg = defaults + cfg

    with open(path, 'w') as f:
        toml.dump(cfg, f)

    return cfg



from qtpy import QtCore, QtWidgets
import qdarkstyle

from pygments import styles

class StyleWidget(QtWidgets.QWidget):
    app_style_changed = QtCore.Signal(str)
    output_style_changed = QtCore.Signal(str)
    source_style_changed = QtCore.Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setup()

    def setup(self):
        self.layout = QtWidgets.QFormLayout()
        self.setLayout(self.layout)

        # workaround: https://github.com/ColinDuquesnoy/QDarkStyleSheet/issues/200
        delegate = QtWidgets.QStyledItemDelegate()
        self.app_combo = QtWidgets.QComboBox()
        self.app_combo.setItemDelegate(delegate)
        self.layout.addRow('Application style', self.app_combo)

        self.output_combo = QtWidgets.QComboBox()
        self.layout.addRow('Output highlight style', self.output_combo)

        self.source_combo = QtWidgets.QComboBox()
        self.layout.addRow('Source highlight style', self.source_combo)

        # add styles
        self.app_combo.addItem('qdarkstyle')
        for style in sorted(QtWidgets.QStyleFactory.keys()):
            self.app_combo.addItem(style.lower())

        for style in sorted(styles.get_all_styles()):
            self.output_combo.addItem(style)
            self.source_combo.addItem(style)

        self.setup_actions()

    def setup_actions(self):
        self.app_combo.currentTextChanged.connect(self.set_app_style)
        self.output_combo.currentTextChanged.connect(self.output_style_changed)
        self.source_combo.currentTextChanged.connect(self.source_style_changed)

    def set_app_style(self, style):
        app = QtWidgets.QApplication.instance()
        if style == 'qdarkstyle':
            app.setStyleSheet(qdarkstyle.load_stylesheet(qt_api='qtpy'))
        else:
            app.setStyleSheet('')
            app.setStyle(style)

        self.app_combo.setCurrentText(style)

    def set_output_style(self, style):
        self.output_combo.setCurrentText(style)

    def set_source_style(self, style):
        self.source_combo.setCurrentText(style)



import os
from setuptools import setup

readme_path = os.path.join(os.path.dirname(__file__), 'README.md')
with open(readme_path) as f:
    readme = f.read()

setup(
    name='telepythy',
    version='0.1.0-1',
    url='https://github.com/dhagrow/telepythy',
    author='Miguel Turner',
    author_email='cymrow@gmail.com',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=['telepythy', 'telepythy.gui'],
    entry_points={
        'console_scripts': [
            'telepythy=telepythy.gui.__main__:main',
            'telepythy-service=telepythy.__main__:main'
            ]
        },
    install_requires=[
        'appdirs',
        'attrdict',
        'Pygments',
        'PySide2',
        'QDarkStyle',
        'QtPy',
        'shiboken2',
        'six',
        'toml',
        ],
    )



from __future__ import print_function, unicode_literals

import prompt_toolkit as pt
from prompt_toolkit.patch_stdout import patch_stdout

from . import logs
from . import sockio
from .utils import start_thread

PS1 = '>>> '
PS2 = '... '

log = logs.get(__name__)

class Client(object):
    def __init__(self, address, timeout=None):
        self._client = sockio.connect(address, timeout)

    def evaluate(self, source):
        self._sendcmd('evaluate', source)

    def complete(self, prefix):
        self._sendcmd('complete', prefix)

    def events(self):
        c = self._client
        self._sendcmd('events')
        while True:
            event = c.recvmsg()

            if event:
                name = event['evt']
                if name == 'done':
                    log.debug('evt: done')
                elif name == 'output':
                    log.debug('out: %r', event['data']['text'][:100])
                else:
                    data = event.get('data', '')
                    data = data and ': ' + repr(data)[:100]
                    log.debug('evt: %s%s', event['evt'], data)

            yield event

    def _sendcmd(self, cmd, data=None):
        msg = {'cmd': cmd}
        if data is not None:
            msg['data'] = data
        log.debug('cmd: %s: %s', cmd, data)
        self._client.sendmsg(msg)

class Repl(object):
    def __init__(self, address):
        self._address = address

        self._session = pt.PromptSession(refresh_interval=0.5)
        self._status = ''

    def repl(self):
        start_thread(self._output)
        client = None

        needs_input = False
        while True:
            try:
                if client is None:
                    client = Client(self._address)
                try:
                    with patch_stdout():
                        source = self._session.prompt(
                            PS2 if needs_input else PS1,
                            bottom_toolbar=lambda: self._status,
                            )

                except KeyboardInterrupt:
                    client.interrupt()
                    needs_input = False
                    print('KeyboardInterrupt')
                    continue
                except EOFError:
                    break

                needs_input = client.evaluate(source)

            except Exception as e:
                client = None
                self._set_disconnected(e)

    def _output(self):
        addr = self._address
        client = None

        while True:
            try:
                if client is None:
                    client = Client(addr, timeout=3)
                for line in client.events():
                    self._set_connected(addr)
                    if line is not None:
                        print(line, end='')
            except Exception as e:
                client = None
                self._set_disconnected(e)

    def _set_connected(self, address):
        msg = '<ansigreen>  </ansigreen> connected: {}:{}'.format(*address)
        self._status = pt.HTML(msg)
        self._session.app.invalidate()

    def _set_disconnected(self, error=None):
        e = error and ': {}'.format(error)
        msg = '<ansired>  </ansired> not connected{}'.format(e)
        self._status = pt.HTML(msg)
        self._session.app.invalidate()

def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-H', '--host', default='localhost')
    parser.add_argument('-P', '--port', required=True)

    args = parser.parse_args()

    Repl((args.host, args.port)).repl()

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass



if __debug__:
    import faulthandler
    faulthandler.enable()

import sys
import signal
import argparse

from qtpy import QtCore, QtGui, QtWidgets

from .. import logs
from .. import control

from . import config
from .window import Window

from .  import resources

def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('-p', '--profile', default='default',
        help='profile or Python executable to load (default: %(default)s)')
    parser.add_argument('-l', '--list-profiles', action='store_true',
        help='list the configured profiles')

    parser.add_argument('--config', default=config.DEFAULT_PATH,
        help='path to the config file (default: %(default)s)')

    parser.add_argument('-q', '--quiet', action='store_true',
        help='disable all output')
    parser.add_argument('-v', '--verbose', action='count', default=0,
        help='enable verbose output (-vv for more)')

    args = parser.parse_args()

    logs.init(args.verbose, args.quiet, mode='ctl')
    cfg = config.init(args.config)

    if args.list_profiles:
        list_profiles(cfg)
        return

    mgr = control.Manager(cfg, args.verbose, args.quiet)

    app = QtWidgets.QApplication()
    # app.setWindowIcon(QtGui.QIcon(':telepathy'))

    win = Window(cfg, mgr, args.profile)
    win.setWindowTitle('Telepythy')
    win.show()

    # enable clean shutdown on ctrl+c
    setup_int_handler(win)

    sys.exit(app.exec_())

def list_profiles(cfg):
    for name, profile in sorted(cfg.profile.items()):
        print(name, profile)

def setup_int_handler(win):
    signal.signal(signal.SIGINT, get_int_handler(win))
    def timer():
        QtCore.QTimer.singleShot(100, timer)
    timer()

def get_int_handler(win):
    def handler(signum, frame):
        win.close()
    return handler

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        print('bye!')



import argparse

from . import logs
from . import utils
from . import service

log = logs.get(__name__)

def main():
    parser = argparse.ArgumentParser('telepythy')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-s', '--serve', nargs='?', default=False,
        help='[default] <interface>:<port> to bind to. '
             'set port to 0 for a random port (default: {})'.format(
            utils.DEFAULT_ADDR))
    group.add_argument('-c', '--connect', nargs='?', default=False,
        help='<host>:<port> to connect to (default: {})'.format(
            utils.DEFAULT_ADDR))

    parser.add_argument('-v', '--verbose', action='count',
        default=0, help='enable verbose output (-vv for more)')
    parser.add_argument('-q', '--quiet', action='store_true',
        help='disable all output')

    args = parser.parse_args()

    logs.init(args.verbose, args.quiet, mode='svc', log_exceptions=False)

    svc = service.Service()

    # serve unless connect is set
    if args.connect is not False:
        addr = utils.parse_address(args.connect or utils.DEFAULT_ADDR)
        svc.connect(addr)
    else:
        addr = utils.parse_address(args.serve or utils.DEFAULT_ADDR)
        svc.serve(addr)

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass



import sys
from logging import *

try:
    import colorlog
except ImportError:
    colorlog = None

LOG_COLORS = {
    'DEBUG'   : 'reset',
    'INFO'    : 'white',
    'WARNING' : 'yellow',
    'ERROR'   : 'red',
    'CRITICAL': 'bold_red,bg_black',
    }

get = getLogger
log = get(__name__)

def init(verbose=0, quiet=False, mode=None, log_exceptions=True):
    """Initializes simple logging defaults."""
    if quiet:
        disable(CRITICAL)
        return

    root_log = get()

    fmt = '%(levelname).1s %(asctime)s [%(mode)s%(name)s] %(message)s'
    if colorlog:
        formatter = colorlog.ColoredFormatter('%(log_color)s' + fmt,
            log_colors=LOG_COLORS)
    else:
        formatter = Formatter(fmt)

    handler = StreamHandler()
    handler.setFormatter(formatter)
    handler.addFilter(ModeFilter(mode))

    root_log.addHandler(handler)
    if   verbose == 0: level = WARNING
    elif verbose == 1: level = INFO
    else:              level = DEBUG
    root_log.setLevel(level)

    if log_exceptions:
        sys.excepthook = handle_exception

    # manage third-party loggers
    get('qdarkstyle').setLevel(WARNING)

def handle_exception(etype, evalue, etb):
    if issubclass(etype, KeyboardInterrupt):
        sys.__excepthook__(etype, evalue, etb)
        return
    log.error('unhandled exception', exc_info=(etype, evalue, etb))

class ModeFilter(Filter):
    def __init__(self, mode):
        super(ModeFilter, self).__init__()
        self._mode = mode + ':' if mode else ''

    def filter(self, record):
        record.mode = self._mode
        return True



import threading

from . import logs

DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 7357
DEFAULT_ADDR = '{}:{}'.format(DEFAULT_HOST, DEFAULT_PORT)

log = logs.get(__name__)

def parse_address(address):
    s = address.split(':', 1)
    host = s[0].strip() or DEFAULT_HOST
    if len(s) == 1 or not s[1]:
        return (host, DEFAULT_PORT)
    return (host, int(s[1]))

def start_thread(func, *args, **kwargs):
    def thread(func, *args, **kwargs):
        ident = threading.current_thread().ident

        log.debug('thread started (%s): %s', ident, func.__name__)
        try:
            return func(*args, **kwargs)
        except:
            log.exception('unexpected thread error')
        finally:
            log.debug('thread stopped (%s): %s', ident, func.__name__)

    t = threading.Thread(target=thread, args=(func,) + args, kwargs=kwargs)
    t.daemon = True
    t.start()
    return t
